;***********************************************************************************
; Mathematikroutinen
;***********************************************************************************


;**************************************************************
; primitive 32 bit Division 	f:= f / xw
;
; output: f  = g = Ganzzahlergebnis
;         r  = Rest der Division
;         xw = unveraendert
; 
;**************************************************************
Div32
	clrf	g0
	decf	g0, f
	clrf	g1
	decf	g1, f
	clrf	g2
	decf	g2, f
	clrf	g3
	decf	g3, f
Div32Loop
	incf	g0, f
	btfsc	STATUS, Z
	incf	g1, f
	btfsc	STATUS, Z
	incf	g2, f
	btfsc	STATUS, Z
	incf	g3, f
	call	Sub32		; 32 bit f:=f-xw   calc=xw cnt=f
	btfss	STATUS, C	; berlauf
	goto	Div32Loop	; Stelle 1 mehr

	; rest der Division sichern nach r
	call	Add32		; f:=f+xw
	movfw	f0
	movwf	r0
	movfw	f1
	movwf	r1
	movfw	f2
	movwf	r2
	movfw	f3
	movwf	r3
	; Divisionsergebnis nach f
	movfw	g0
	movwf	f0
	movfw	g1
	movwf	f1
	movfw	g2
	movwf	f2
	movfw	g3
	movwf	f3

	return


;*******************************************************************************
;Multiplikation: 40-Bit x 2 (BIN4)
Mal2
	bcf	STATUS,C
	rlf	BIN4, f
	rlf	BIN4+1, f
	rlf	BIN4+2, f
	rlf	BIN4+3, f
	rlf	BIN4+4, f
	return


;*******************************************************************************
;Division: 40-Bit / 2 (BIN4)
Durch2
	bcf	STATUS,C
	rrf	BIN4+4, f
	rrf	BIN4+3, f
	rrf	BIN4+2, f
	rrf	BIN4+1, f
	rrf	BIN4, f
	return


;*******************************************************************************
;Multiplikation: 40-Bit x 5 (BIN4)
Mal5
	movfw	BIN4
	movwf	xw0
	movfw	BIN4+1
	movwf	xw1
	movfw	BIN4+2
	movwf	xw2
	movfw	BIN4+3
	movwf	xw3
	movfw	BIN4+4
	movwf	xw4
	call	Mal2		; x2
	call	Mal2		; x4
				; 32-bit add: BIN4 = BIN4 + xw
	movf	xw0,W		; low byte
	addwf	BIN4,F 		; low byte add

	movf	xw1,W 		; next byte
	btfsc	STATUS,C 	; berspringe falls C nicht gesetzt
	incfsz	xw1,W 		; addiere C falls gesetzt 
	addwf	BIN4+1,F	; next byte add wenn NZ
	
	movf	xw2,W 		; next byte
	btfsc	STATUS,C 	; berspringe falls C nicht gesetzt
	incfsz	xw2,W 		; addiere C falls gesetzt 
	addwf	BIN4+2,F 	; next byte add wenn NZ
	
	movf	xw3,W 		; high byte
	btfsc	STATUS,C 	; berspringe falls C nicht gesetzt
	incfsz	xw3,W 		; addiere C falls gesetzt 
	addwf	BIN4+3,F	; high byte add wenn NZ
	
	movf	xw4,W 		; high byte
	btfsc	STATUS,C 	; berspringe falls C nicht gesetzt
	incfsz	xw4,W 		; addiere C falls gesetzt 
	addwf	BIN4+4,F	; high byte add wenn NZ

	return


;*******************************************************************************
;Multiplikation: 40-Bit x 40 (BIN)  40 = 2x2x2x5
Mal40
	call	Mal2		;  2 : 1
	call	Mal2		;  4 : 1
	call	Mal2		;  8 : 1
	call	Mal5		; 40 : 1
	return

;*******************************************************************************
;Multiplikation: 40-Bit x 64 (BIN)
Mal64
	call	Mal2		;  2 : 1
	call	Mal2		;  4 : 1
	call	Mal2		;  8 : 1
	call	Mal2		; 16 : 1
	call	Mal2		; 32 : 1
	call	Mal2		; 64 : 1
	return

;*******************************************************************************
;Multiplikation: 16-Bit x 40 (f)  40 = 2x2x2x5
Mal40f
	call	Mal2f		;  2 : 1
	call	Mal2f		;  4 : 1
	call	Mal2f		;  8 : 1
	clrf	f2
	clrf	f3
	call	Mal5f32		; 40 : 1
	return

;*******************************************************************************
;Multiplikation: 16-Bit x 64 (f)
Mal64f
	call	Mal2f		;  2 : 1
	call	Mal2f		;  4 : 1
	call	Mal2f		;  8 : 1
	call	Mal2f		; 16 : 1
	call	Mal2f		; 32 : 1
	call	Mal2f		; 64 : 1
	return

;*******************************************************************************
;Multiplikation: 16-Bit x 2 (f)
Mal2f
	bcf	STATUS,C
	rlf	f0, f
	rlf	f1, f
	return

;*******************************************************************************
;Multiplikation: 32-Bit x 2 (f)
Mal2f32
	bcf	STATUS,C
	rlf	f0, f
	rlf	f1, f
	rlf	f2, f
	rlf	f3, f
	return

;*******************************************************************************
;Multiplikation: 32-Bit x 5 (f)
Mal5f32
	movfw	f0
	movwf	xw0
	movfw	f1
	movwf	xw1
	movfw	f2
	movwf	xw2
	movfw	f3
	movwf	xw3
	call	Mal2f32
	call	Mal2f32
	call	Add32
	return

;*******************************************************************************
;Multiplikation: 32-Bit x 1000 (f)
Mal1000f32
	call	Mal5f32
	call	Mal5f32
	call	Mal5f32
	call	Mal2f32
	call	Mal2f32
	call	Mal2f32
	return

;*******************************************************************************
;Division: 16-Bit /2 (f)
Durch2f
	bcf	STATUS,C
	rrf	f1, f
	rrf	f0, f
	return

;*******************************************************************************
;Division: 16-Bit /4 (f)
Durch4f
	bcf	STATUS,C
	rrf	f1, f
	rrf	f0, f
	bcf	STATUS,C
	rrf	f1, f
	rrf	f0, f
	return


;**************************************************************
; Mathematikroutinen
;***********************************************************
; 16 Bit Wert (f1,f0) 5-stellig auf LCD dezimal anzeigen mit Leadin 0
OutDez16L0			;16-bit (f1,f0) als Dez zum Lcd
	call	Hex2Dez16	;16 Bit to BCD
	clrf	Fehler		; Leading0 vorbereiten
	goto	OutDez32halbe


;***********************************************************
; 32 Bit Wert (f3,f2,f1,f0) 10-stellig auf LCD dezimal anzeigen
OutDez32			;32-bit (f3,f2,f1,f0) als Dez zum Lcd
	call	Hex2Dez32	;32 Bit to BCD

	clrf	Fehler		; Leading0 vorbereiten

	movfw	HdG
	Call	Leading0	; 1 xxx xxx xxx 

	movlw	' '
	btfsc	PpmMode
	movlw	'*'		; kennzeichnet Praezisionsmode
	btfsc	PpmMode2
	movlw	'#'		; kennzeichnet Praezisionsmode2
	movwf	LcdDaten
	call	OutLcdDaten

	movfw	HdHM	
	Call	Leading0	; 1 1xx xxx xxx 
	movfw	HdZM
	Call	Leading0	; 1 11x xxx xxx 
	btfsc	Pulsmode6
	bsf	Fehler,0	; lesding0 abschalten bei 6 nachkommastellen
	movfw	HdM
	Call	Leading0	; 1 111 xxx xxx 

	movlw	' '
	btfsc	Pulsmode6
	movlw	'.'
	movwf	LcdDaten
	call	OutLcdDaten

	movfw	HdHT
	Call	Leading0	; 1 111 1xx xxx 
	movfw	HdZT
	Call	Leading0	; 1 111 11x xxx 

OutDez32halbe
	btfsc	Pulsmode3
	bsf	Fehler,0	; lesding0 abschalten bei 3 nachkommastellen

	movfw	HdT
	Call	Leading0	; 1 111 111 xxx 

	movlw	' '
	btfsc	Pulsmode3
	movlw	'.'
	movwf	LcdDaten
	btfss	DCF_dppm
	call	OutLcdDaten

	movfw	HdH
	Call	Leading0	; 1 111 111 1xx 
	movfw	HdZ
	Call	Leading0	; 1 111 111 11x 

	movlw	','
	movwf	LcdDaten
	btfsc	DCF_dppm
	call	OutLcdDaten

	movfw	HdE
	Call	Bcd4Bit		; 1 111 111 111 
	return

Leading0
	iorwf	Fehler, f
	movf	Fehler, f	; Test auf 0
	btfss	STATUS, Z	; bisher alles '0'
	goto	Bcd4Bit		; nein
	movlw	' '		; ja
	movwf	LcdDaten
	goto	OutLcdDaten


;**************************************************************
; 16 Bit Wert (f1,f0) auf LCD dezimal anzeigen
OutDez16			;16-bit (f0,f1) als Dez zum Lcd
	call	Hex2Dez16
	movfw	HdZT
	Call	Bcd4Bit
	movfw	HdT
	Call	Bcd4Bit

	movlw	' '
	movwf	LcdDaten

	movfw	HdH
	Call	Bcd4Bit
	movfw	HdZ
	Call	Bcd4Bit
	movfw	HdE
	Call	Bcd4Bit
	return


;**************************************************************
; 32-bit(f3,f2,f1,f0) in 10-stellen Bcd (G,HM,ZM,M,HT,ZT,T,H,Z,E)

;1 000 000 000 = 3B9A CA00 h
;  100 000 000 = 05F5 E100 h
;   10 000 000 = 0098 9680 h
;    1 000 000 = 000F 4240 h
;      100 000 = 0001 86A0 h
;       10 000 = 0000 2710 h
;        1 000 = 0000 03E8 h
;          100 = 0000 0064 h
;           10 = 0000 000A h
;            1 = 0000 0001 h

Hex2Dez32			; 32-bit(f3,f2,f1,f0) in 10-stellen Bcd (G,HM,ZM,M,HT,ZT,T,H,Z,E)

	movlw	0x3B		; 1 000 000 000 = 3B 9A CA 00 h	
	movwf	xw3
	movlw	0x9A		
	movwf	xw2
	movlw	0xCA
	movwf	xw1
	movlw	0x00
	movwf	xw0
	call	Hex2Dez1	; 1 000 000 000er
	movfw	HdX
	movwf	HdG

	movlw	0x05		; 100 000 000 = 05 F5 E1 00 h	
	movwf	xw3
	movlw	0xF5		
	movwf	xw2
	movlw	0xE1
	movwf	xw1
	movlw	0x00
	movwf	xw0
	call	Hex2Dez1	; 100 000 000er
	movfw	HdX
	movwf	HdHM

	clrf	xw3		; 10 000 000 = 00 98 96 80 h
	movlw	0x98		
	movwf	xw2
	movlw	0x96
	movwf	xw1
	movlw	0x80
	movwf	xw0
	call	Hex2Dez1	; 10 000 000er
	movfw	HdX
	movwf	HdZM

	clrf	xw3		; 1 000 000 = 00 0F 42 40 h
	movlw	0x0F		
	movwf	xw2
	movlw	0x42
	movwf	xw1
	movlw	0x40
	movwf	xw0
	call	Hex2Dez1	; 1 000 000er
	movfw	HdX
	movwf	HdM

	clrf	xw3		; 100 000 = 00 01 86 A0 h
	movlw	0x01		
	movwf	xw2
	movlw	0x86
	movwf	xw1
	movlw	0xA0
	movwf	xw0
	call	Hex2Dez1	; 100 000er
	movfw	HdX
	movwf	HdHT
Hex2Dez16			; 16-bit(f1,f0) in 5-stellen Bcd (ZT,T,H,Z,E)				
	movlw	0x27		; 10 000 = 00 00 27 10 h
	movwf	xw1
	clrf	xw2
	clrf	xw3
	movlw	0x10
	movwf	xw0
	call	Hex2Dez1	; 10 000er
	movfw	HdX
	movwf	HdZT

	movlw	0x03		; 1 000 = 00 00 03 E8 h
	movwf	xw1
	clrf	xw2
	clrf	xw3
	movlw	0xE8
	movwf	xw0
	call	Hex2Dez1	; 1000er
	movfw	HdX
	movwf	HdT
Hex2Dez8			; 8-bit (f1, f0) in 3-stellen BCD
	movlw	0x00		; 100 = 00 00 00 64 h
	movwf	xw3
	movwf	xw2
	movwf	xw1
	movlw	0x64
	movwf	xw0
	call	Hex2Dez1	; 100er
	movfw	HdX
	movwf	HdH

	movlw	0x00		; 10 = 00 00 00 0A h
	movwf	xw3
	movwf	xw2
	movwf	xw1
	movlw	0x0A
	movwf	xw0
	call	Hex2Dez1	; 10er
	movfw	HdX
	movwf	HdZ

	movfw	f0
	movwf	HdE
	return

Hex2Dez1
	clrf	HdX
	decf	HdX, f
HdLoop
	incf	HdX, f
	call	Sub32		;
	btfss	STATUS, C	;berlauf
	goto	HdLoop		;Stelle 1 mehr
	call	Add32
	return


;**************************************************************
; 32 Bit Subtraktion, bei berlauf (neg. Ergebnis) ist C gesetzt
Sub32				; 32 bit f:=f-xw   calc=xw cnt=f
	clrf	Fehler		; extraflags lschen    
	movf    xw0, w		; f0=f0-xw0
	subwf   f0, f

	btfsc   STATUS,C
	goto    sb0
	movlw   0x01		; borgen von f1
	subwf   f1, f

	btfsc   STATUS,C
	goto    sb0
	subwf   f2    ,f        ; borgen von f2

	btfsc   STATUS,C
	goto    sb0
	subwf   f3    ,f        ; borgen von f3
	btfss   STATUS,C
	bsf	Fehler, C	; unterlauf

sb0     movf    xw1, w		; f1=f1-xw1
	subwf   f1, f

	btfsc   STATUS,C
	goto    sb1
	movlw   0x01		; borgen von f2
	subwf   f2    ,f

	btfsc   STATUS,C
	goto    sb1
	subwf   f3, f		; borgen von f3

        btfss   STATUS,C
	bsf	Fehler, C	; Unterlauf

sb1     movf    xw2,w		; f2=f2-xw2
        subwf   f2, f

        btfsc   STATUS,C
        goto    sb2
        movlw   0x01		
        subwf   f3, f		; borgen von f3

        btfss   STATUS,C
       	bsf	Fehler, C	; Unterlauf

sb2     movf    xw3,w		; f3=f3-xw3
        subwf   f3    ,f

        btfss   STATUS,C
 	bsf	Fehler, C	; Unterlauf

	bcf	STATUS, C
	btfsc	Fehler, C
	bsf	STATUS, C
	return


;32 bit Adition, C-Flag bei berlauf gesetzt 
;********************************************************************* 
;*  
;* addiere xw (32-byte value) mit f (32-byte value). 
;* Input  variablen: f (32-bit value), xw (32-bit value) 
;* Output variablen: f (32-bit value), 33. bit in STATUS,C 
;********************************************************************* 
Add32 				; 32-bit add: f = f + xw
	movf	xw0,W		; low byte
	addwf	f0,F 		; low byte add

	movf	xw1,W 		; next byte
	btfsc	STATUS,C 	; berspringe falls C nicht gesetzt
	incfsz	xw1,W 		; addiere C falls gesetzt 
	addwf	f1,F 		; next byte add wenn NZ
	
	movf	xw2,W 		; next byte
	btfsc	STATUS,C 	; berspringe falls C nicht gesetzt
	incfsz	xw2,W 		; addiere C falls gesetzt 
	addwf	f2,F 		; next byte add wenn NZ
	
	movf	xw3,W 		; high byte
	btfsc	STATUS,C 	; berspringe falls C nicht gesetzt
	incfsz	xw3,W 		; addiere C falls gesetzt 
	addwf	f3,F 		; high byte add wenn NZ
	
	return 			; fertig


;**************************************************************
;8-Bit als BCD 2-stellig ausgeben
OutBcd				;8-Bit als BCD ausgeben
	movwf	BcdDaten	;8-Bit kopieren
	movwf	BcdDaten1
	rrf	BcdDaten, f
	rrf	BcdDaten, f
	rrf	BcdDaten, f
	rrf	BcdDaten, f	;4 High-Bit in BcdDaten
	movlw	B'00001111'
	andwf	BcdDaten, f	; / 
	andwf	BcdDaten1, f	; 4-Low-Bit in BcdDaten1
	movfw	BcdDaten
	call	Bcd4Bit		;4 High-Bit ausgeben
	movfw	BcdDaten1
	call	Bcd4Bit		;4 Low-Bit ausgeben
	return
Bcd4Bit				;low-4 Bit als BCD ausgeben
	movwf	BcdDaten
	movlw	B'00110000'
	ADDwf	BcdDaten, f	;ASCII-wandeln (+48)
	movlw	B'00111010'
	subwf	BcdDaten, w
	btfss	STATUS, C	;Test auf A ... F
	goto	BcdOk
	movlw	.7
	addwf	BcdDaten, f	;korrigiere A...F (+7)
BcdOk
	movfw	BcdDaten
	call	OutLcdDaten
	return



;*****************************************************************	
;
; Zeitverzoegerungen
;
;*****************************************************************	
;*****************************************************************	
;Zeitverzgerung um 200 ms
Wait200ms
	movlw	D'200'
	movwf	loops
	goto	WAIT		; 200ms 

;*****************************************************************	
;Zeitverzgerung um 100 ms
Wait100ms
	movlw	D'100'
	movwf	loops
	goto	WAIT		; 100ms 

;*****************************************************************	
;Zeitverzgerung um 50 ms
Wait50ms
	movlw	D'50'
	movwf	loops
	goto	WAIT		; 50ms 

;*****************************************************************	
;Zeitverzgerung um 20 ms
Wait20ms
	movlw	D'20'
	movwf	loops
	goto	WAIT		; 20ms 

;*****************************************************************	
;Zeitverzgerung um 10 ms
Wait10ms
	movlw	D'10'
	movwf	loops
	goto	WAIT		; 10ms 

;*****************************************************************	
;Zeitverzgerung um 5 ms
Wait5ms
	movlw	D'5'
	movwf	loops
	goto	WAIT		; 5ms 

;*****************************************************************	
;Zeitverzgerung um loops * 1 ms
;20MHz ->   5MHz -> 5000 zyklen

WAIT
top      movlw   .249           ; timing adjustment variable (1ms)  20MHz 249*20=4980 zyklen
         movwf   loops2
top2     nop                     ; sit and wait (Schleifenlnge 20T)
         nop
         nop
         nop
         nop
         nop
	 nop
         nop
         nop
         nop
         nop
         nop
         nop
         nop
         nop
         nop
         nop
         decfsz  loops2, F      ; inner loops complete?
         goto    top2           ; no, go again
                                ;
         decfsz  loops, F       ; outer loops complete?
         goto    top            ; no, go again
         retlw   0              ; yes, return from subWAIT

;*****************************************************************	